/****************************************************************************
 *
 * ACTIVE SILICON LIMITED
 *
 * File name   : WaterDemo.cxx
 * Function    : Example application showing Interopebility with CUDA/OpenGL
 * Project     : Phoenix
 *
 * Copyright (c) 2018 Active Silicon Ltd.
 ****************************************************************************
 * Comments:
 * --------
 * This example shows a shallow water Lattice Boltzman model used as a 
 * refraction surface overlayed on top of the given camera's live feed.
 *
 ****************************************************************************
 */

// OpenGL Graphics includes
// Windows
#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
#define WINDOWS_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#define USE_TEXSUBIMAGE2D
#pragma warning(disable:4996)
#include "glad/glad.h"
// Apple MacOSX
#elif defined(__APPLE__) || defined(MACOSX)
#include <OpenGL/gl3.h>
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
#define USE_TEXSUBIMAGE2D
#define __IS_APPLE__
// Linux
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "helper_gl.h"

// STL includes
#include <vector>

// CUDA includes
#include <cuda_runtime.h>
#include "cuda_gl_interop.h"

// CUDA helper include files from CUDA 7.5 SDK
#include "helper_cuda.h"
#include "helper_cuda_gl.h"

#if defined(__APPLE__) || defined(MACOSX)
#define GLFW_INCLUDE_GLCOREARB
#define GLFW_INCLUDE_GLEXT
#include <GLFW/glfw3.h>
#else
#include "GLFW/glfw3.h"
#endif

#include "WaterLBComputation.h"
#include "Random.h"

//-------------------------------------------------------------------------------------------

GLuint compileGLSLprogram(const char *vertex_shader_src, const char *fragment_shader_src)
{
    GLuint v, f, p;

	p = glCreateProgram();

	if(vertex_shader_src != 0)
	{
		v = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(v, 1, &vertex_shader_src, NULL);
		glCompileShader(v);

		// check if shader compiled
		GLint compiled = 0;
		glGetShaderiv(v, GL_COMPILE_STATUS, &compiled);

		if(!compiled)
		{
			char temp[256] = "";
			glGetShaderInfoLog(v, 256, NULL, temp);
			fprintf(stderr, "Vtx Compile failed:\n%s\n", temp);
			glDeleteShader(v);
			return 0;
		}
		else
		{
			glAttachShader(p,v);
		}
	}

	if(fragment_shader_src != 0)
	{
		f = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(f, 1, &fragment_shader_src, NULL);
		glCompileShader(f);

		// check if shader compiled
		GLint compiled = 0;
		glGetShaderiv(f, GL_COMPILE_STATUS, &compiled);

		if(!compiled)
		{
			char temp[256] = "";
			glGetShaderInfoLog(f, 256, NULL, temp);
			fprintf(stderr, "frag Compile failed:\n%s\n", temp);
			glDeleteShader(f);
			return 0;
		}
		else
		{
			glAttachShader(p,f);
		}
	}

	glLinkProgram(p);

	int infologLength = 0;
	int charsWritten  = 0;

	glGetProgramiv(p, GL_INFO_LOG_LENGTH, reinterpret_cast<GLint *>(&infologLength));

	if(infologLength > 1)
	{
		char *infoLog = reinterpret_cast<char *>(malloc(infologLength));
		glGetProgramInfoLog(p, infologLength, reinterpret_cast<GLsizei *>(&charsWritten), infoLog);
		fprintf(stderr, "Shader compilation error: %s\n", infoLog);
		free(infoLog);
	}

	return p;
}

//-------------------------------------------------------------------------------------------

static const char *glsl_vertex_shader_src = 
	"#version 410 core\n"
	"layout(location = 0) in vec2 vVertex;\n"
	"smooth out vec2 vUV;\n"
	"void main()\n"
	"{\n"
	"    gl_Position = vec4(vVertex * 2.0 - 1.0, 0.0, 1.0);\n"
	"    vUV = vVertex;\n"
	"}\n";

//-------------------------------------------------------------------------------------------

static const char *glsl_fragment_tex_shader_src = 
	"#version 410 core\n"
	"layout(location = 0) out vec4 vFragColor;\n"
	"smooth in vec2 vUV;\n"
	"uniform usampler2D textureMap;\n"
	"void main() {\n"
	"    vec2 uv = vec2(vUV.x, 1.0 - vUV.y);\n"
	"    vec4 t = texture(textureMap, uv).rbga;\n"
	"    vFragColor = t / 255.0;\n"
	"}\n";


//-------------------------------------------------------------------------------------------

static const char *glsl_fragment_pbo_shader_src = 
	"#version 410 core\n"
	"layout(location = 0) out vec4 vFragColor;\n"
	"smooth in vec2 vUV;\n"
	"uniform sampler2D textureMap;\n"
	"void main()\n"
	"{\n"
	"    vec2 uv = vec2(vUV.x, 1.0 - vUV.y);\n"
	"    vec4 t = texture(textureMap, uv).rbga;\n"
	"    vFragColor = t;\n"
	"}\n";


//-------------------------------------------------------------------------------------------

class WaterDemo
{
	public:
		virtual ~WaterDemo();
		
		static WaterDemo *instance(int argc, char **argv, bool usePBOMethod);
		static WaterDemo *instance();

		void paint() const;
		void iterate();
		static void resizeCB(GLFWwindow *window, int width, int height);

	private:
	
		static WaterDemo *m_instance;

		/** Number of grid points along the x-axis of the water computation grid.
		  * In order to fit the CUDA algorithm this value must be a multiple of 16.
		  * Also the width * height must be a multiple of 576 to fit the CUDA algorithm.
		  */
		static const int c_waterGridWidth = 1344 * 2;

		/** Number of grid points along the y-axis of the water computation grid.
		  * In order to fit the CUDA algorithm this value must be a multiple of 16.
		  * Also the width * height must be a multiple of 576 to fit the CUDA algorithm.
		  */		
		static const int c_waterGridHeight = 768 * 2;
		
		/** Two methods of rendering are available dependent upon the platform.
		  * The PBO method uses a Pixel Buffer Object to transfer image data from
		  * CUDA to OpenGL's pipeline.
		  * The texture method uses texture buffer to transfer image data from CUDA
		  * to OpenGL's pipeline.
		  * */
		bool m_usePBOMethod;
				
		/** Handle to computation of Lattice-Boltzman shallow water model.
		  * */
		WaterLBComputeData *m_waterModel;
		int m_step;
		
		/** Handle to the OpenGL shader program.
		 * */
		GLuint m_program;

		GLuint m_vaoID;
		GLuint m_vboVerticesID;
		// For texture based rendering
		GLuint m_textureId;
		// For PBO based rendering
		GLuint m_pboId;

		WaterDemo(bool usePBOMethod);
		
		bool isPBOMethod() const;
		
		bool init(int argc, char **argv);
		void cleanup();
		
		bool setupVBO();
		bool setupTexture();
		bool setupPBO();
		
		void bringTheRain() const;		
};

//-------------------------------------------------------------------------------------------

WaterDemo *WaterDemo::m_instance = 0;

//-------------------------------------------------------------------------------------------

WaterDemo::WaterDemo(bool usePBOMethod) : m_usePBOMethod(usePBOMethod),
	m_waterModel(0),
	m_step(0),
	m_program(0),
	m_vaoID(0),
	m_vboVerticesID(0),
	m_textureId(0),
	m_pboId(0)
{}

//-------------------------------------------------------------------------------------------

WaterDemo::~WaterDemo()
{
	cleanup();
	m_instance = 0;
}

//-------------------------------------------------------------------------------------------

WaterDemo *WaterDemo::instance(int argc, char **argv, bool usePBOMethod)
{
	if(m_instance == 0)
	{
		m_instance = new WaterDemo(usePBOMethod);
		if(!m_instance->init(argc,argv))
		{
			delete m_instance;
			m_instance = 0;
		}
	}
	return m_instance;
}

//-------------------------------------------------------------------------------------------

WaterDemo *WaterDemo::instance()
{
	return m_instance;
}

//-------------------------------------------------------------------------------------------

bool WaterDemo::isPBOMethod() const
{
	return m_usePBOMethod;
}

//-------------------------------------------------------------------------------------------

bool WaterDemo::init(int argc, char **argv)
{
	bool res = false;
	
	// If no device found then method calls exit
	if(findCudaGLDevice(argc, const_cast<const char **>(argv)) < 0)
	{
		return false;
	}
	
	m_waterModel = waterLBAllocateAndInit(c_waterGridWidth, c_waterGridHeight);
	if(m_waterModel != 0)
	{
		m_step = 0;
		
		if(isPBOMethod())
		{
			m_program = compileGLSLprogram(glsl_vertex_shader_src, glsl_fragment_pbo_shader_src);
		}
		else
		{
			m_program = compileGLSLprogram(glsl_vertex_shader_src, glsl_fragment_tex_shader_src);
		}
		if(m_program != 0)
		{
			if(setupVBO())
			{
				res = true;
			}
		}
		else
		{
			fprintf(stderr, "Failed to compile required OpenGL GLSL shaders.\n");
		}
	}
	else
	{
		fprintf(stderr, "Failed to allocate CUDA LB water model on GPU.\n");
	}
	return res;
}

//-------------------------------------------------------------------------------------------

void WaterDemo::cleanup()
{
	if(m_vboVerticesID != 0)
	{
		glDeleteBuffers(1, &m_vboVerticesID);
		m_vboVerticesID = 0;
	}
	if(m_vaoID != 0)
	{	
		glDeleteVertexArrays(1, &m_vaoID);
		m_vaoID = 0;
	}	
	if(m_textureId != 0)
	{
		glDeleteTextures(1, &m_textureId);
		m_textureId = 0;
	}
	if(m_pboId != 0)
	{
		glDeleteBuffers(1, &m_pboId);
		m_pboId = 0;
	}
	if(m_waterModel != 0)
	{
		waterLBFree(m_waterModel);
		m_waterModel = 0;
	}
}

//-------------------------------------------------------------------------------------------

bool WaterDemo::setupVBO()
{
	std::vector<float> vertices;

	if(!setupTexture())
	{
		return false;
	}

	// bottom left = A
	vertices.push_back(0.0);
	vertices.push_back(0.0);
	// bottom right = B
	vertices.push_back(1.0);
	vertices.push_back(0.0);
	// top right = D
	vertices.push_back(1.0);
	vertices.push_back(1.0);
	// bottom left = A
	vertices.push_back(0.0);
	vertices.push_back(0.0);
	// top right = D
	vertices.push_back(1.0);
	vertices.push_back(1.0);
	// top left = C
	vertices.push_back(0.0);
	vertices.push_back(1.0);
	
	GLint vVertexLocation = glGetAttribLocation(m_program, "vVertex");

	glGenVertexArrays(1, &m_vaoID);
	glBindVertexArray(m_vaoID);

	glGenBuffers(1, &m_vboVerticesID);
	glBindBuffer(GL_ARRAY_BUFFER, m_vboVerticesID);
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(GLfloat), &vertices[0], GL_STATIC_DRAW);

	glEnableVertexAttribArray(0);
	glVertexAttribPointer(vVertexLocation, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(GL_FLOAT), 0);
	glEnableVertexAttribArray(vVertexLocation);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return IS_SDK_CHECK_OK_GL();
}

//-------------------------------------------------------------------------------------------

bool WaterDemo::setupTexture()
{
	if(isPBOMethod())
	{
		if(!setupPBO())
		{
			return false;
		}
	}
	
	glGenTextures(1, &m_textureId);
	glBindTexture(GL_TEXTURE_2D, m_textureId);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	
	if(isPBOMethod())
	{
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, c_waterGridWidth, c_waterGridHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}
	else
	{
#if !defined(USE_TEXSUBIMAGE2D)
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8UI_EXT, c_waterGridWidth, c_waterGridHeight, 0, GL_RGBA_INTEGER_EXT, GL_UNSIGNED_BYTE, 0);
#endif
		// Associate OpenGL texture with CUDA
		if(!waterLBSetupTextureOutput(m_waterModel, m_textureId))
		{
			fprintf(stderr,"Failed to associated texture with CUDA-OpenGL interoperability\n");
			return false;
		}
	}
	
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return IS_SDK_CHECK_OK_GL();
}

//-------------------------------------------------------------------------------------------

bool WaterDemo::setupPBO()
{
	unsigned int nTexels, nValues, nBytes;
	
	// Based on CUDA/OpenGL Interop example, where memory of the PBO is allocated.
	nTexels = c_waterGridWidth * c_waterGridHeight;
	nValues = nTexels * 4;
	nBytes = sizeof(GLubyte) * nValues;
	void *data = ::malloc(nBytes);
	
	if(data == 0)
	{
		return false;
	}
	
	glGenBuffers(1, &m_pboId);
	glBindBuffer(GL_ARRAY_BUFFER, m_pboId);
	glBufferData(GL_ARRAY_BUFFER, nBytes, data, GL_DYNAMIC_DRAW);
	::free(data);
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	
	// Associate OpenGL PBO pixel buffer with CUDA
	if(!waterLBSetupPBOOutput(m_waterModel, m_pboId))
	{
		fprintf(stderr, "Failed to associated PBO with CUDA-OpenGL interoperability\n");
		return false;
	}
	
	return IS_SDK_CHECK_OK_GL();
}

//-------------------------------------------------------------------------------------------

void WaterDemo::paint() const
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glUseProgram(m_program);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, m_textureId);
	GLint tLocation = glGetUniformLocation(m_program, "textureMap");
	glUniform1i(tLocation, 0);
	
	glBindVertexArray(m_vaoID);
	glBindBuffer(GL_ARRAY_BUFFER, m_vboVerticesID);
	glDrawArrays(GL_TRIANGLES, 0, 6);
}

//-------------------------------------------------------------------------------------------

void WaterDemo::resizeCB(GLFWwindow *window, int width, int height)
{
	glViewport(0, 0, width, height);
}

//-------------------------------------------------------------------------------------------

void WaterDemo::iterate()
{
	// Run the water model
	waterLBCollisionProcess3(m_waterModel, (m_step>0) ? true : false);
	waterLBStream(m_waterModel);
	waterLBSwap(m_waterModel);
	waterLBIncrementForce(m_waterModel);
	bringTheRain();
	m_step++;
	
	// Fill the approriate GL destination with result
	waterLBRenderHeightTexture(m_waterModel);
	
	if(isPBOMethod())
	{
		glBindTexture(GL_TEXTURE_2D, m_textureId);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, m_pboId);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, c_waterGridWidth, c_waterGridHeight, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
	}
}

//-------------------------------------------------------------------------------------------

void WaterDemo::bringTheRain() const
{
	const float c_rainFrequency = 0.1f;
	
	Random *rand = Random::instance();
	float maxRadius = static_cast<float>(c_waterGridWidth) / 30.0f;
	int loopCount = 0;
	
	while(rand->randomReal1() < c_rainFrequency && loopCount<4)
	{
		int posX = static_cast<int>(rand->randomUInt32() % c_waterGridWidth);
		int posY = static_cast<int>(rand->randomUInt32() % c_waterGridHeight);
		float rForce = 0.0002f;
		float rRadius = maxRadius;
		waterLBRaindropCircleForce(m_waterModel, static_cast<float>(posX), static_cast<float>(posY), rForce, rRadius);
		loopCount++;
	}	
}

//-------------------------------------------------------------------------------------------

/*
int main(int argc, char **argv)
{
#if defined(USE_TEXSUBIMAGE2D)
	bool usePBO = true;
#else
	bool usePBO = false;
#endif
	
	WaterDemo *demo = WaterDemo::instance(argc, argv, usePBO);
	if(demo != 0)
	{
		delete demo;
	}
	return 0;
}
*/

//-------------------------------------------------------------------------------------------

int main(int argc, char **argv)
{
#if defined(USE_TEXSUBIMAGE2D)
	bool usePBO = true;
#else
	bool usePBO = false;
#endif

	GLFWwindow *window;

	if(glfwInit() != GLFW_TRUE)
	{
		return -1;
	}

#if defined(__APPLE__) || defined(MACOSX)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 2);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
#endif

	// Create a windowed mode window and its OpenGL context
	window = glfwCreateWindow(640, 480, "Hello World", NULL, NULL);
	if(window == 0)
	{
		glfwTerminate();
		return -1;
	}

	glfwSetWindowSizeCallback(window, WaterDemo::resizeCB);

	// Make the window's context current
	glfwMakeContextCurrent(window);
#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
	if(!gladLoadGLLoader(reinterpret_cast<GLADloadproc>(glfwGetProcAddress)))
	{
		fprintf(stderr, "Failed to initialize OpenGL context\n");
		glfwTerminate();
		return -1;
	}
#endif
	glViewport(0, 0, 640, 480);

	WaterDemo *demo = WaterDemo::instance(argc, argv, usePBO);
	if(demo != 0)
	{
		// Loop until the user closes the window
		while(!glfwWindowShouldClose(window))
		{
			demo->iterate();
			demo->paint();
			// Swap front and back buffers
			glfwSwapBuffers(window);
			// Poll for and process events
			glfwPollEvents();
		}
	}

	glfwTerminate();
	return 0;
}
